
/*
BMP2DATA.CPP - Converts bitmaps in 8-24 bit color into C/ASM tile code for XGS AVR/PIC systems

Author: Andre' Lamothe
Last Modified: 1.2.09
Notes: 

Description:  This program us used to convert a set of bitmap tiles that are typically
8x8 into 256 color C/ASM code that are compatible with the tile mode graphics drivers for
the XGS AVR/PIC, the program has to take in the tile bitmaps, convert to 24-bit color first
then extract the bitmap(s), then for each color in the image find the nearest match in the 
target NTSC 8-bit color words OR converts to RGB 2.2.2 8-bit format for VGA output.

This is a very nastly program and not pretty, like all tools its a "work in progress" and 
coding on the fly with complete disregard to anything :), the idea is to get it done and 
make it work. In other words its a "travesty" of coding which is typical of tools since 
in most cases they start out as a few lines of main() console code, then you add functions, 
then you copy things, etc. so the point is don't worry about messy tools, no one is going 
to look at the code, they just need to work!

The program requires this palette file to be in the working directory:

    xgs_default_NTSC_palette_192x1.bmp

The book gives specific examples of using the tool, here is its formal usage:

Usage BMP2DATA.exe inputfile [flags] > outputfile");
flags = -B -D -VGA -Axx -TW[1...255] -TH[1..255] -W[1..256] -H[1..256] -C[1..256] -STx,y -XYx,y -FX -FY -V -I -P[0|1] -?");

-B    = Enable 1 pixel border for templated bitmaps, default no border.");
-D    = Output data in decimal format, default is hex.")
-VGA  = Perform conversion in VGA RGB space to 2.2.2 format, default is NTSC format color words.")
-Axx  = Alpha mapper, maps xx 8-bit color index to 0, this is useful in NTSC modes where $4F is "black", but sprites need 0 as "transparent".);
-TWxx = Width of tile set on x-axis, default = 1.");
-THxx = Height of tile set on y-axis, default = 1.");
-Wxx  = Width of tile, default = 8 pixels.");
-Hxx  = Height of tile, default = 8 pixels.");
-Cxx  = Total count of tiles to be converted, if omitted assumed to be tw*th.");
-STxx,yy = Starting tile x,y indices to start scan from in scan mode, default 0,0
-XYxx,yy = Coordinate of single tile to pull from bitmap, upper left (0,0) overrides -Cxx.");
-FX   = Flips the output bitmap's on the X axis.");
-FY   = Flips the output bitmap on the Y axis.";
-P file = Loads in another palette bmp file, must be 192x1x256 BMP format."); 
-I    = Enables interactive mode.");
-V    = Verbose flag.");
-?    = Prints help.");

Examples: The following examples highlight a number of common scenarios you might encounter when ripping tiles and how you would 
command the tool to accomplish the task.

For the examples we will use the file "sprites_8x8_01.bmp" which is a tiled bitmap of images each 8x8, with a 1 pixel border. 
the "template" that the tiles are drawn in is 8 columns and 20 rows. The rows isn't as important and the columns since the tool
need the rows to compute the step to return to the next row for ripping.


Example 1: Extract 8 tiles from the 0th row in NTSC mode with no transparency, that is black is encoded as NTSC black $4F (79 decimal)

    BMP2DATA.exe sprites_8x8_01.bmp -B -C8 -TW8 -TH8

    Explanation: This command line says to rip from sprites_8x8_01.bmp, use a border "-B", count off 8 tiles "-C8", the width and height of the
    tile map template is 8 columns "-TW8" by 8 rows "-TH8" (ignore the remaining rows).

Example 2: Extract 4 tiles from the 0th row, starting at tile 4,0 in NTSC mode with no transparency, that is black is encoded as NTSC black $4F (79 decimal)

    BMP2DATA.exe sprites_8x8_01.bmp -B -C4 -TW8 -TH8 -ST4,0

    Explanation: This command line is identical to the last except that it has two changes, first we only want 4 tiles ripped, so "-C4", 
    secondly, we want the ripping to start from (4,0), "-ST4,0".

Example 3: Redo example 1, but output VGA compatible data.

    BMP2DATA.exe sprites_8x8_01.bmp -B -C8 -TW8 -TH8 -VGA

    Explanation: This command line says to rip from sprites_8x8_01.bmp, use a border "-B", count off 8 tiles "-C8", the width and height of the
    tile map template is 8 columns "-TW8" by 8 rows "-TH8" (ignore the remaining rows), lastly "-VGA" instructs the tool to output VGA formatted
    RGB data rather than try and match NTSC palette entries and outout LUMA:CHROMA data.

Example 4: Say you want to extract a sigle tile from column 2, row 3, (2,3) in NTSC mode from a template with a border, 8x8 tiles:

    BMP2DATA.exe sprites_8x8_01.bmp -B -TW8 -TH8 -XY2,3

Example 5:  Now say that you want to redo example 4, but this time, you are ripping a tile to be used as a sprite, thus "black" you want
            mapped to 0, so the driver makes it transparent. 

    BMP2DATA.exe sprites_8x8_01.bmp -B -TW8 -TH8 -XY2,3 -A79

    Explanation: the trailing "-A79", enables the "alpha mapper" which find any code and maps it to 0. In this case, it was directed to use
    79 decimal which is $4F hex. This maps the NTSC "black" code to 0 for use as a sprite. Remember, sprites need 0 where you want transparent.
    In VGA output modes, this is not needed since RGB black happens to be 0. However, if you wanted to map a DIFFERENT RGB code maybe pink to
    0, then you could use the alpha flag.

Example 6: Rip the entire bmp into tiles

    BMP2DATA.exe sprites_8x8_01.bmp -B -C144 -TW8 -TH8

    Explanation: This rips the entire bitmap into data, all 144 tiles, NTSC mode, 8x8 tiles as always is default.

*/


// INCLUDES /////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#define INITGUID

#include <windows.h>   // include important windows stuff
#include <windowsx.h>
#include <mmsystem.h>
#include <conio.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <io.h>
#include <fcntl.h>
#include <sys/timeb.h>
#include <time.h>

// DEFINE ///////////////////////////////////////////////////////////////

// directdraw pixel format defines, used to help
// bitmap loader put data in proper format
#define DD_PIXEL_FORMAT8        8
#define DD_PIXEL_FORMAT555      15
#define DD_PIXEL_FORMAT565      16
#define DD_PIXEL_FORMAT888      24
#define DD_PIXEL_FORMATALPHA888 32

#define MAX_COLORS_PALETTE  256

// bitmap defines
#define BITMAP_ID            0x4D42 // universal id for a bitmap


// conversion modes for bitmaps
#define XGS_NTSC 0 // default mode
#define XGS_PAL  1 // (not supported yet)
#define XGS_VGA  2 // VGA format

#define LUMA_BLACK 4  // this is the value the luma words are "biased" or offset by to match the video signal 
                      // and form proper [LUMA4:CHROMA4] data


// TYPES ////////////////////////////////////////////////////////////////
// basic unsigned types
typedef unsigned short USHORT;
typedef unsigned short WORD;
typedef unsigned char  UCHAR;
typedef unsigned char  BYTE;
typedef unsigned int   QUAD;
typedef unsigned int   UINT;

// container structure for bitmaps .BMP file
typedef struct BITMAP_FILE_TAG
        {
        BITMAPFILEHEADER bitmapfileheader;  // this contains the bitmapfile header
        BITMAPINFOHEADER bitmapinfoheader;  // this is all the info including the palette
        PALETTEENTRY     palette[256];      // we will store the palette here
        UCHAR            *buffer;           // this is a pointer to the data

        } BITMAP_FILE, *BITMAP_FILE_PTR;

// the simple bitmap image
typedef struct BITMAP_IMAGE_TYP
        {
        int state;          // state of bitmap
        int attr;           // attributes of bitmap
        int x,y;            // position of bitmap
        int width, height;  // size of bitmap
        int num_bytes;      // total bytes of bitmap
        int bpp;            // bits per pixel
        UCHAR *buffer;      // pixels of bitmap

        } BITMAP_IMAGE, *BITMAP_IMAGE_PTR;

// structure to hold xgs palette entry in a couple formats to help color analysis algorithm, basically just a 4-tuple named array
typedef struct XGS_PALETTE_ENTRY_TYP
    {
    int colors[4];
    } XGS_PALETTE_ENTRY, *XGS_PALETTE_ENTRY_PTR;


#pragma pack(1) // this is needed so that word alignment doesn't bloat the data structure, since it needs to be EXACTLY 18 bytes long!
// TGA header
typedef struct TGA_HEADER_TYP
{
    BYTE identsize;          // size of ID field that follows 18 byte header (0 usually)
    BYTE colourmaptype;      // type of colour map 0=none, 1=has palette
    BYTE imagetype;          // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

    short colourmapstart;     // first colour map entry in palette
    short colourmaplength;    // number of colours in palette
    BYTE colourmapbits;      // number of bits per palette entry 15,16,24,32

    short xstart;             // image x origin
    short ystart;             // image y origin
    short width;              // image width in pixels
    short height;             // image height in pixels
    BYTE bits;               // image bits per pixel 8,16,24,32
    BYTE descriptor;         // image descriptor bits (vh flip bits)
    
    // pixel data follows header....
    
} TGA_HEADER, *TGA_HEADER_PTR; 
#pragma pack()


/*
--------------------------------------------------------------------------------
DATA TYPE 2:  Unmapped RGB images.                                             |
_______________________________________________________________________________|
| Offset | Length |                     Description                            |
|--------|--------|------------------------------------------------------------|
|--------|--------|------------------------------------------------------------|
|    0   |     1  |  Number of Characters in Identification Field.             |
|        |        |                                                            |
|        |        |  This field is a one-byte unsigned integer, specifying     |
|        |        |  the length of the Image Identification Field.  Its value  |
|        |        |  is 0 to 255.  A value of 0 means that no Image            |
|        |        |  Identification Field is included.                         |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    1   |     1  |  Color Map Type.                                           |
|        |        |                                                            |
|        |        |  This field contains either 0 or 1.  0 means no color map  |
|        |        |  is included.  1 means a color map is included, but since  |
|        |        |  this is an unmapped image it is usually ignored.  TIPS    |
|        |        |  ( a Targa paint system ) will set the border color        |
|        |        |  the first map color if it is present.                     |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    2   |     1  |  Image Type Code.                                          |
|        |        |                                                            |
|        |        |  This field will always contain a binary 2.                |
|        |        |  ( That's what makes it Data Type 2 ).                     |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    3   |     5  |  Color Map Specification.                                  |
|        |        |                                                            |
|        |        |  Ignored if Color Map Type is 0; otherwise, interpreted    |
|        |        |  as follows:                                               |
|        |        |                                                            |
|    3   |     2  |  Color Map Origin.                                         |
|        |        |  Integer ( lo-hi ) index of first color map entry.         |
|        |        |                                                            |
|    5   |     2  |  Color Map Length.                                         |
|        |        |  Integer ( lo-hi ) count of color map entries.             |
|        |        |                                                            |
|    7   |     1  |  Color Map Entry Size.                                     |
|        |        |  Number of bits in color map entry.  16 for the Targa 16,  |
|        |        |  24 for the Targa 24, 32 for the Targa 32.                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    8   |    10  |  Image Specification.                                      |
|        |        |                                                            |
|    8   |     2  |  X Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) X coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   10   |     2  |  Y Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) Y coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   12   |     2  |  Width of Image.                                           |
|        |        |  Integer ( lo-hi ) width of the image in pixels.           |
|        |        |                                                            |
|   14   |     2  |  Height of Image.                                          |
|        |        |  Integer ( lo-hi ) height of the image in pixels.          |
|        |        |                                                            |
|   16   |     1  |  Image Pixel Size.                                         |
|        |        |  Number of bits in a pixel.  This is 16 for Targa 16,      |
|        |        |  24 for Targa 24, and .... well, you get the idea.         |
|        |        |                                                            |
|   17   |     1  |  Image Descriptor Byte.                                    |
|        |        |  Bits 3-0 - number of attribute bits associated with each  |
|        |        |             pixel.  For the Targa 16, this would be 0 or   |
|        |        |             1.  For the Targa 24, it should be 0.  For     |
|        |        |             Targa 32, it should be 8.                      |
|        |        |  Bit 4    - reserved.  Must be set to 0.                   |
|        |        |  Bit 5    - screen origin bit.                             |
|        |        |             0 = Origin in lower left-hand corner.          |
|        |        |             1 = Origin in upper left-hand corner.          |
|        |        |             Must be 0 for Truevision images.               |
|        |        |  Bits 7-6 - Data storage interleaving flag.                |
|        |        |             00 = non-interleaved.                          |
|        |        |             01 = two-way (even/odd) interleaving.          |
|        |        |             10 = four way interleaving.                    |
|        |        |             11 = reserved.                                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|   18   | varies |  Image Identification Field.                               |
|        |        |  Contains a free-form identification field of the length   |
|        |        |  specified in byte 1 of the image record.  It's usually    |
|        |        |  omitted ( length in byte 1 = 0 ), but can be up to 255    |
|        |        |  characters.  If more identification information is        |
|        |        |  required, it can be stored after the image data.          |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Color map data.                                           |
|        |        |                                                            |
|        |        |  If the Color Map Type is 0, this field doesn't exist.     |
|        |        |  Otherwise, just read past it to get to the image.         |
|        |        |  The Color Map Specification describes the size of each    |
|        |        |  entry, and the number of entries you'll have to skip.     |
|        |        |  Each color map entry is 2, 3, or 4 bytes.                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Image Data Field.                                         |
|        |        |                                                            |
|        |        |  This field specifies (width) x (height) pixels.  Each     |
|        |        |  pixel specifies an RGB color value, which is stored as    |
|        |        |  an integral number of bytes.                              |
|        |        |                                                            |
|        |        |  The 2 byte entry is broken down as follows:               |
|        |        |  ARRRRRGG GGGBBBBB, where each letter represents a bit.    |
|        |        |  But, because of the lo-hi storage order, the first byte   |
|        |        |  coming from the file will actually be GGGBBBBB, and the   |
|        |        |  second will be ARRRRRGG. "A" represents an attribute bit. |
|        |        |                                                            |
|        |        |  The 3 byte entry contains 1 byte each of blue, green,     |
|        |        |  and red.                                                  |
|        |        |                                                            |
|        |        |  The 4 byte entry contains 1 byte each of blue, green,     |
|        |        |  red, and attribute.  For faster speed (because of the     |
|        |        |  hardware of the Targa board itself), Targa 24 images are  |
|        |        |  sometimes stored as Targa 32 images.                      |
|        |        |                                                            |
--------------------------------------------------------------------------------
*/



// MACROS ///////////////////////////////////////////////////////////////

// this builds a 16 bit color value in 5.5.5 format (1-bit alpha mode)
#define _RGB16BIT555(r,g,b) ((b & 31) + ((g & 31) << 5) + ((r & 31) << 10))

// this builds a 16 bit color value in 5.6.5 format (green dominate mode)
#define _RGB16BIT565(r,g,b) ((b & 31) + ((g & 63) << 5) + ((r & 31) << 11))

// this builds a 24 bit color value in 8.8.8 format
#define _RGB24BIT(a,r,g,b) ((b) + ((g) << 8) + ((r) << 16) )

// this builds a 32 bit color value in A.8.8.8 format (8-bit alpha mode)
#define _RGB32BIT(a,r,g,b) ((b) + ((g) << 8) + ((r) << 16) + ((a) << 24))

// bit manipulation macros
#define SET_BIT(word,bit_flag)   ((word)=((word) | (bit_flag)))
#define RESET_BIT(word,bit_flag) ((word)=((word) & (~bit_flag)))

// used to compute the min and max of two expresions
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define SIGN(a) ( ((a) > 0) ? (1) : (-1) )

// used for swapping algorithm
#define SWAP(a,b,t) {t=a; a=b; b=t;}

// some math macros
#define DEG_TO_RAD(ang) ((ang)*PI/180.0)
#define RAD_TO_DEG(rads) ((rads)*180.0/PI)

#define RAND_RANGE(x,y) ( (x) + (rand()%((y)-(x)+1)))

// PROTOTYPES ///////////////////////////////////////////////////////////

int Load_Bitmap_File(BITMAP_FILE_PTR bitmap, char *filename, int flip=1);
int Load_Bitmap_File2(BITMAP_FILE_PTR bitmap, char *filename, int flip=1);
int Flip_Bitmap(UCHAR *image, int bytes_per_line, int height);
int Extract_Bitmap(int tile_x, int tile_y, int width, int height, int step_x, int step_y, BYTE *bitmap_buffer, BITMAP_FILE_PTR bitmap);
int Build_Palette_From_Bitmap(int tx, int ty, int width, int height, BYTE *bitmap_buffer, int output_bitmap_flag);

// EXTERNALS ////////////////////////////////////////////////////////////


// GLOBALS //////////////////////////////////////////////////////////////

int cmd_B  = 0;  // Enable 1 pixel border templated bitmaps, default no border.
int cmd_D  = 0;  // radix, 1 = decimal, 0=hex, default
int cmd_VGA = XGS_NTSC; // default is NTSC mode
int cmd_A  = -1; // alpha control, default do nothing, if non-zero then as bitmap is converted then this code xx will be mapped to 0 always.
int cmd_TW = 1;  // width of tile set on x-axis, default = 1.
int cmd_TH = 1;  // height of tile set on y-axis, default = 1.
int cmd_W  = 8;  // width of tile, default = 8.
int cmd_H  = 8;  // height of tile, default = 8.
int cmd_C  = 1;  // total count of tiles to be converted, if omitted assumed to bee tw*th
int cmd_XY = 0;  // flags if x,y should be used for single bitmap extraction
int cmd_SX = 0;  // tile starting x for scan modes    
int cmd_SY = 0;  // tile starting y for scan modes    
int cmd_X  = 0;  // coordinate of single tile to pull from bitmap, overrides -Cxx
int cmd_Y  = 0; 
int cmd_FX = 0; // Flips the output bitmap's on the X axis (needed for hel engine 4.0)
int cmd_FY = 0; // Flips the output bitmap on the Y axis
int cmd_I  = 0; // enables interactive mode
int cmd_V  = 0; // verbose flag

char cmd_P[32]  = ""; // palette select, null is the default

char inputfilename[256]; // input file to converter
char inputfileroot[256]; // input file root without extention

char cstring1[256], cstring2[256]; // working strings
char *cstring_ptr1, *cstring_ptr2; // working string pointers
int cindex1, cindex2; // general search indices

BITMAP_FILE bitmap;                 // a n bit bitmap file
BITMAP_FILE palette_bitmap;         // holds the palette bitmap (192x1), for NTSC mode conversion
BITMAP_FILE bitmap8bit;             // a 8 bit bitmap file
BITMAP_FILE bitmap16bit;            // a 16 bit bitmap file
BITMAP_FILE bitmap24bit;            // a 24 bit bitmap file

char		filename[256];			// general filename
char        sbuffer[256];           // working string
int dd_pixel_format = DD_PIXEL_FORMAT565;    // default pixel format

int num_xgs_colors = 192; // number of colors in xgs color map

float r_bias = 0, g_bias = 0, b_bias = 0;
float r_scale = 1.1, g_scale = 0.6, b_scale = 1.1; 

// the xgs color look up map, use PALETTEENTRY to hold r,g,b and flags field is xgs "color" word

PALETTEENTRY *xgs_color_map = NULL;


// the xgs color look up map, use PALETTEENTRY to hold r,g,b and flags field is xgs "color" word
PALETTEENTRY xgs_color_map0[256] = 
{
// scan this out of BMP file from now on, left to right over for each color 0..15, shade 4..15, 16*12=192 colors 
};

PALETTEENTRY xgs_color_map1[256] = 
{
// scan this out of BMP file from now on, left to right over for each color 0..15, shade 4..15, 16*12=192 colors 
};


// FUNCTIONS ////////////////////////////////////////////////////////////

int Extract_Bitmap(int tile_x, int tile_y, // x,y tile location 
                   int width, int height,  // size of tile to extract (default 8x8)
                   int step_x, int step_y, // step to move from tile to tile, eg. 8,8 no border, 9,9 with 1-pixel border
                   BYTE *bitmap_buffer,    // destination buffer for bitmap data (width*height*3)
                   BITMAP_FILE_PTR bitmap) // pointer to the bitmap structure itself holding pertinent bitmap info
{
// this function extracts a bitmap tile from the large bitmap image

// step 1: locate exact bitmap coordinates to extract from
int xb = (tile_x * step_x) + (step_x-width);
int yb = (tile_y * step_y) + (step_y-height);

// compute starting pixel address to perform copy from
BYTE *source_bitmap_addr = (BYTE *)bitmap->buffer + 3*(xb + yb * bitmap->bitmapinfoheader.biWidth);
BYTE *dest_buffer = bitmap_buffer;

// now copy the pixels from the source to destination
for (int y=0; y < height; y++)
    {
    // copy pixels
    memcpy(dest_buffer, source_bitmap_addr, width*3);

#if 0    
    printf("\ncopying line %d, src_addr = %p, dest_addr = %p\n", y, (void *)source_bitmap_addr, (void *)dest_buffer);
    for (int x=0; x < width*3; x++)
        {
        printf("%d,", source_bitmap_addr[x]);
        dest_buffer [x] = source_bitmap_addr[x];

        } // end for index
    printf("\n");
#endif
 
    // advance pointers
    dest_buffer        += (width*3);
    source_bitmap_addr += (bitmap->bitmapinfoheader.biWidth*3);    
    } // end for y


#if 0
// write TGA file out to confirm image
TGA_HEADER tga;

// open file, if doesn't exist throw error and exit
FILE *fp = fopen("bitmap2.tga", "wb");

// build TGA header
tga.identsize       = 0;   // size of ID field that follows 18 byte header (0 usually)
tga.colourmaptype   = 0;   // type of colour map 0=none, 1=has palette
tga.imagetype       = 2;   // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

tga.colourmapstart  = 0;    // first colour map entry in palette
tga.colourmaplength = 0;    // number of colours in palette
tga.colourmapbits   = 0;   // number of bits per palette entry 15,16,24,32

tga.xstart          = 0;    // image x origin
tga.ystart          = 0;    // image y origin
tga.width           = bitmap->bitmapinfoheader.biWidth;     // image width in pixels
tga.height          = bitmap->bitmapinfoheader.biHeight;    // image height in pixels
tga.bits            = bitmap->bitmapinfoheader.biBitCount;  // image bits per pixel 8,16,24,32
tga.descriptor      = 0x30; // image descriptor bits (vh flip bits)

 
// write descriptor out
printf("\nWriting TGA header size = %d", sizeof(tga));
fwrite((void *)&tga, sizeof(tga), 1, fp);

printf("\nTGA data size = %d", bitmap->bitmapinfoheader.biSizeImage);
// write the bitmap data out
fwrite((void *)bitmap->buffer, bitmap->bitmapinfoheader.biSizeImage, 1, fp);

// close the file
fclose(fp);

#endif


#if 0
// write TGA file out to confirm image
TGA_HEADER tga;

// build a filename
sprintf(filename,"bitmap_x%d_y%d_w%d_h%d_sx%d_sy%d.tga", tile_x, tile_y, width, height, step_x, step_y);

// open file, if doesn't exist throw error and exit
FILE *fp = fopen(filename, "wb");

// build TGA header
tga.identsize       = 0;    // size of ID field that follows 18 byte header (0 usually)
tga.colourmaptype   = 0;    // type of colour map 0=none, 1=has palette
tga.imagetype       = 2;    // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

tga.colourmapstart  = 0;    // first colour map entry in palette
tga.colourmaplength = 0;    // number of colours in palette
tga.colourmapbits   = 0;    // number of bits per palette entry 15,16,24,32

tga.xstart          = 0;    // image x origin
tga.ystart          = 0;    // image y origin
tga.width           = width;     // image width in pixels
tga.height          = height;    // image height in pixels
tga.bits            = 24;    // image bits per pixel 8,16,24,32
tga.descriptor      = 0x30;  // image descriptor bits (vh flip bits)

 
// write descriptor out
printf("\nWriting TGA header size = %d", sizeof(tga));
fwrite((void *)&tga, sizeof(tga), 1, fp);

printf("\nTGA data size = %d", width*height*3);
// write the bitmap data out
fwrite((void *)bitmap_buffer, width*height*3, 1, fp);

// close the file
fclose(fp);

#endif


// return success
return(1);


} // end Extract_Bitmap

/////////////////////////////////////////////////////////////////////////


int Convert_Bitmap_To_Palette(BYTE *input_bitmapRGB,  // input bitmap in RGB format 
                              int width, int height,  // size of bitmap to convert
                              int tx, int ty,         // tile indices from caller, send (0,0) if single tile
                              int mode,               // conversion mode XGS_NTSC, XGS_PAL (not supported yet), XGS_VGA
                              BYTE *output_bitmap)    // output bitmap in 8-bit "data" word format ready for the XGS drivers
                                                      // each 8-bit value is in NTSC, PAL, or VGA (RGB mode)
{
XGS_PALETTE_ENTRY match_pal;   // the best match color palette constructed, holds indices to map at first
int num_match_pal_colors = 0;  // the number of colors inserted into the match palette
PALETTEENTRY src_col;          // color map compatible palette used to perform searching algorithm

int px, py;                 // indexing variables
BYTE r,g,b;                 // used to extract red, green, blue components of pixel
int index, index2;          // general looping index


int col_dist = 256*256*3;   // used in color matching algorithm to compute distance to color in 3D colorspace
int best_col_match = -1;    // index of best match of color in color matching algorithm

int r_target, g_target, b_target;
int r_test, g_test, b_test;

UCHAR chroma, luma;        // used in conversion process


// first the easy case...
if (mode == XGS_VGA)
    {
    // simply divide each RGB value down and merge into a single 6-bit 2.2.2 format word
for (py = 0; py < height; py++)
    {

if (cmd_V) 
   {
   printf("\n");
   } // end if 

    for (px = 0; px < width; px++)
        {
        // extract pixel at (px, py) and get RGB, shift down 6-bit -- ouch!!!
        r_target = input_bitmapRGB[3*(px + py*width) + 0] >> 6;
        g_target = input_bitmapRGB[3*(px + py*width) + 1] >> 6;
        b_target = input_bitmapRGB[3*(px + py*width) + 2] >> 6;

        // write data in proper format RGB - msb [B2.G2.R2] lsb
        output_bitmap[px + py*width ] =  (b_target << 4) | (g_target << 2) | (r_target);
        } // end for px

    } // end for py

    } // end if 
else
if (mode == XGS_NTSC || XGS_PAL) // give pal its own conversion later
{

// for every pixel convert to target XGS palette
for (py = 0; py < height; py++)
    {

if (cmd_V) 
   {
   printf("\n");
   } // end if 

    for (px = 0; px < width; px++)
        {
        // extract pixel at (px, py) and get RGB
        r_target = input_bitmapRGB[3*(px + py*width) + 0];
        g_target = input_bitmapRGB[3*(px + py*width) + 1];
        b_target = input_bitmapRGB[3*(px + py*width) + 2];

if (cmd_V) 
   {
   printf("\nSearching at p(%d, %d) for [%d, %d, %d]",px, py, r_target, g_target, b_target );        
   } // end if

        // reset color distance to max
        col_dist       = 256*256*3;
        best_col_match = -1; 

        // for each pixel in bitmap match to colors in final palette
        for (index = 0; index < num_xgs_colors; index++)
            {
            // extract test r,g,b we are testing against for match?
            r_test = xgs_color_map[ index ].peRed;
            g_test = xgs_color_map[ index ].peGreen;       
            b_test = xgs_color_map[ index ].peBlue;
                        
            // compare test rgb to target rgb and if its closer update best match
            int rgb_dist = (r_target - r_test) * (r_target - r_test) + 
                           (g_target - g_test) * (g_target - g_test) + 
                           (b_target - b_test) * (b_target - b_test);
        
            if (rgb_dist < col_dist)
               {
               // update distance and index 
               col_dist       = rgb_dist; 
               best_col_match = index;
               } // end if

            } // end for index
  
        // at this point col_dist and best_col_match are the best matches for the pixel in question, ready to write it out! FINALLY!
if (cmd_V) 
   {
   printf(" | Found best match at palette entry %2d, ",best_col_match);        
   } // end if

        // convert index to chroma luma data with correct offset, etc.
        chroma =               best_col_match / 12;  // 16 colors, 12 shades per bank
        luma   = LUMA_BLACK + (best_col_match % 12); // 12 shades for each color, bias luma by BLACK = 4


        // write data in proper format [LUMA4: CHROMA4] format
        output_bitmap[px + py*width ] = (luma << 4) | (chroma);
            
        // test if alpha mapper is on?
        if (cmd_A != -1)
            {
            // if this pixel is the alpha target then map it to 0, which is used by all engines as the "transparent" value.
            if ( output_bitmap[px + py*width ] == cmd_A)
                 output_bitmap[px + py*width ] = 0;
            } // end else

if (cmd_V) 
   {
   printf("| Best match converted to NTSC format = %X, ",output_bitmap[px + py*width ]);        
   } // end if

        } // end for pixel_x

    } // end for pixel_y

if (cmd_V) 
   {
   printf("\n\n");
   } // end if 

} // end if NTSC/PAL modes



// we have everything we need now to output the XGS C/C++ driver compliant code 

#if 1
// loop thru image and output C/C++/ASM data definitions code
printf("\n\n// Tile extracted from file \"%s\" at tile=(%d, %d), size=(%d, %d)", inputfilename, 
                                                                                 tx, ty, 
                                                                                 width, height );
for (py = 0; py < height; py++)
    {
    printf("\n");

    for (px = 0; px < width; px++)
        {
        // compute final indices taking into consideration mirroring/flipping flags
        int xi = px;
        int yi = py;
        
        if (cmd_FX) 
           xi = (width-1) - xi;

        if (cmd_FY) 
           yi = (height-1) - yi;

        // test output mode
        if (cmd_D == 0)
            printf("0x%02X,",output_bitmap[xi + yi*width ]);
        else // decimal
            printf("%d,",output_bitmap[xi + yi*width ]);

        } // end for px

    } // end for py

printf("\n");

#endif


// return index of color palette in master list
return(1);

} // end Convert_Bitmap_To_Palette

/////////////////////////////////////////////////////////////////////////

int Load_Bitmap_File2(BITMAP_FILE_PTR bitmap, char *filename, int flip)
{
// this function opens a bitmap file and loads the data into bitmap
// it will convert 8, 16, 24 bit bitmaps all into 24-bit data, so 
// palette matching to xgs palette has the most accuracy possible

int file_handle,  // the file handle
    index;        // looping index

UCHAR   *temp_buffer = NULL; // used to convert images to 24 bit
OFSTRUCT file_data;          // the file data information

// open the file if it exists
if ((file_handle = OpenFile(filename,&file_data,OF_READ))==-1)
   return(0);

// now load the bitmap file header
_lread(file_handle, &bitmap->bitmapfileheader,sizeof(BITMAPFILEHEADER));

// test if this is a bitmap file
if (bitmap->bitmapfileheader.bfType!=BITMAP_ID)
   {
   // close the file
   _lclose(file_handle);

   // return error
   return(0);
   } // end if

// now we know this is a bitmap, so read in all the sections

// first the bitmap infoheader

// now load the bitmap file header
_lread(file_handle, &bitmap->bitmapinfoheader,sizeof(BITMAPINFOHEADER));

// compute size of image and store back in structure, some BMP tools mess this up
bitmap->bitmapinfoheader.biSizeImage = bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight*bitmap->bitmapinfoheader.biBitCount/8;

// now load the color palette if there is one
if (bitmap->bitmapinfoheader.biBitCount == 8)
   {
   _lread(file_handle, &bitmap->palette,MAX_COLORS_PALETTE*sizeof(PALETTEENTRY));

   // now set all the flags in the palette correctly and fix the reversed
   // BGR RGBQUAD data format
   for (index=0; index < MAX_COLORS_PALETTE; index++)
       {
       // reverse the red and blue fields
       int temp_color                = bitmap->palette[index].peRed;
       bitmap->palette[index].peRed  = bitmap->palette[index].peBlue;
       bitmap->palette[index].peBlue = temp_color;

       // always set the flags word to this
       bitmap->palette[index].peFlags = PC_NOCOLLAPSE;
       } // end for index

    } // end if

// now, seek to the image data itself
 _lseek(file_handle,-(int)(bitmap->bitmapinfoheader.biSizeImage),SEEK_END);

// now read in the image for 8-bit case
if (bitmap->bitmapinfoheader.biBitCount==8)
   {
   // delete the last image if there was one
   if (bitmap->buffer)
       free(bitmap->buffer);

   // allocate the memory for the 8-bit image
   if (!(temp_buffer = (UCHAR *)malloc(bitmap->bitmapinfoheader.biSizeImage)))
      {
      // close the file
      _lclose(file_handle);

      // return error
      return(0);
      } // end if

   // allocate final 24 bit storage buffer
   if (!(bitmap->buffer=(UCHAR *)malloc(3*bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight)))
      {
      // close the file
      _lclose(file_handle);

      // release working buffer
      free(temp_buffer);

      // return error
      return(0);
      } // end if

   // now read it in to temp file
   _lread(file_handle,temp_buffer,bitmap->bitmapinfoheader.biSizeImage);

   // start the conversion process from 8bit to 24bit
   for (index=0; index < bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight; index++)
       {
        // look up color in palette, compute RGB values
        UCHAR red   = bitmap->palette[ temp_buffer[index] ].peRed;        
        UCHAR green = bitmap->palette[ temp_buffer[index] ].peGreen;
        UCHAR blue  = bitmap->palette[ temp_buffer[index] ].peBlue;
        
       // write colors back out to 24-bit buffer
       ((UCHAR *)bitmap->buffer)[index*3 + 0] = red;
       ((UCHAR *)bitmap->buffer)[index*3 + 1] = green;
       ((UCHAR *)bitmap->buffer)[index*3 + 2] = blue;

       } // end for index

   // finally write out the correct number of bits
   bitmap->bitmapinfoheader.biBitCount=24;

   // recomputing bitmap size
   bitmap->bitmapinfoheader.biSizeImage = bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight*bitmap->bitmapinfoheader.biBitCount/8;

   // release working buffer
   free(temp_buffer);

   } // end if
else
if (bitmap->bitmapinfoheader.biBitCount==24)
   {
   // allocate temporary buffer to load 24 bit image
   if (!(temp_buffer = (UCHAR *)malloc(bitmap->bitmapinfoheader.biSizeImage)))
      {
      // close the file
      _lclose(file_handle);

      // return error
      return(0);
      } // end if

   // allocate final 24 bit storage buffer
   if (!(bitmap->buffer=(UCHAR *)malloc(3*bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight)))
      {
      // close the file
      _lclose(file_handle);

      // release working buffer
      free(temp_buffer);

      // return error
      return(0);
      } // end if

   // now read the file in
   _lread(file_handle,temp_buffer,bitmap->bitmapinfoheader.biSizeImage);

   // now convert each 24 bit BGR value into 24 bit RGB value, simple byte swaping
   for (index=0; index < bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight; index++)
       {
           // extract RGB components (in BGR order)
        UCHAR blue  = temp_buffer[index*3 + 0];
        UCHAR green = temp_buffer[index*3 + 1];
        UCHAR red   = temp_buffer[index*3 + 2];
   
       // write colors back out to buffer
       ((UCHAR *)bitmap->buffer)[index*3 + 0] = red;
       ((UCHAR *)bitmap->buffer)[index*3 + 1] = green;
       ((UCHAR *)bitmap->buffer)[index*3 + 2] = blue;

       } // end for index

   // finally write out the correct number of bits (Redundant)
   bitmap->bitmapinfoheader.biBitCount=24;

   // release working buffer
   free(temp_buffer);

   } // end if 24 bit
else
   {
   // serious problem
   return(0);

   } // end else

if (cmd_V) {
// write the file info out
printf("\nfilename:%s \nsize=%d \nwidth=%d \nheight=%d \nbitsperpixel=%d \ncolors=%d \nimpcolors=%d\n",
        filename,
        bitmap->bitmapinfoheader.biSizeImage,
        bitmap->bitmapinfoheader.biWidth,
        bitmap->bitmapinfoheader.biHeight,
		bitmap->bitmapinfoheader.biBitCount,
        bitmap->bitmapinfoheader.biClrUsed,
        bitmap->bitmapinfoheader.biClrImportant);
}

// close the file
_lclose(file_handle);

// flip the bitmap
if (flip)
    Flip_Bitmap(bitmap->buffer,
                bitmap->bitmapinfoheader.biWidth*(bitmap->bitmapinfoheader.biBitCount/8),
                bitmap->bitmapinfoheader.biHeight);


#if 0
// write TGA file out to confirm image
TGA_HEADER tga;

// open file, if doesn't exist throw error and exit
FILE *fp = fopen("bitmap2.tga", "wb");

// build TGA header
tga.identsize       = 0;   // size of ID field that follows 18 byte header (0 usually)
tga.colourmaptype   = 0;   // type of colour map 0=none, 1=has palette
tga.imagetype       = 2;   // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

tga.colourmapstart  = 0;    // first colour map entry in palette
tga.colourmaplength = 0;    // number of colours in palette
tga.colourmapbits   = 0;   // number of bits per palette entry 15,16,24,32

tga.xstart          = 0;    // image x origin
tga.ystart          = 0;    // image y origin
tga.width           = bitmap->bitmapinfoheader.biWidth;     // image width in pixels
tga.height          = bitmap->bitmapinfoheader.biHeight;    // image height in pixels
tga.bits            = bitmap->bitmapinfoheader.biBitCount;  // image bits per pixel 8,16,24,32
tga.descriptor      = 0x30; // image descriptor bits (vh flip bits)

 
// write descriptor out
printf("\nWriting TGA header size = %d", sizeof(tga));
fwrite((void *)&tga, sizeof(tga), 1, fp);

printf("\nTGA data size = %d", bitmap->bitmapinfoheader.biSizeImage);
// write the bitmap data out
fwrite((void *)bitmap->buffer, bitmap->bitmapinfoheader.biSizeImage, 1, fp);

// close the file
fclose(fp);

#endif


// return success
return(1);

} // end Load_Bitmap_File2


/////////////////////////////////////////////////////////////////////////

int Load_Bitmap_File(BITMAP_FILE_PTR bitmap, char *filename, int flip)
{
// this function opens a bitmap file and loads the data into bitmap

int file_handle,  // the file handle
    index;        // looping index

UCHAR   *temp_buffer = NULL; // used to convert 24 bit images to 16 bit
OFSTRUCT file_data;          // the file data information

// open the file if it exists
if ((file_handle = OpenFile(filename,&file_data,OF_READ))==-1)
   return(0);

// now load the bitmap file header
_lread(file_handle, &bitmap->bitmapfileheader,sizeof(BITMAPFILEHEADER));

// test if this is a bitmap file
if (bitmap->bitmapfileheader.bfType!=BITMAP_ID)
   {
   // close the file
   _lclose(file_handle);

   // return error
   return(0);
   } // end if

// now we know this is a bitmap, so read in all the sections

// first the bitmap infoheader

// now load the bitmap file header
_lread(file_handle, &bitmap->bitmapinfoheader,sizeof(BITMAPINFOHEADER));

// compute size of image and store back in structure, some BMP tools mess this up
bitmap->bitmapinfoheader.biSizeImage = bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight*bitmap->bitmapinfoheader.biBitCount/8;

// now load the color palette if there is one
if (bitmap->bitmapinfoheader.biBitCount == 8)
   {
   _lread(file_handle, &bitmap->palette,MAX_COLORS_PALETTE*sizeof(PALETTEENTRY));

   // now set all the flags in the palette correctly and fix the reversed
   // BGR RGBQUAD data format
   for (index=0; index < MAX_COLORS_PALETTE; index++)
       {
       // reverse the red and green fields
       int temp_color                = bitmap->palette[index].peRed;
       bitmap->palette[index].peRed  = bitmap->palette[index].peBlue;
       bitmap->palette[index].peBlue = temp_color;

       // always set the flags word to this
       bitmap->palette[index].peFlags = PC_NOCOLLAPSE;
       } // end for index

    } // end if


// finally the image data itself
   _lseek(file_handle,-(int)(bitmap->bitmapinfoheader.biSizeImage),SEEK_END);

// now read in the image
if (bitmap->bitmapinfoheader.biBitCount==8 || bitmap->bitmapinfoheader.biBitCount==16)
   {
   // delete the last image if there was one
   if (bitmap->buffer)
       free(bitmap->buffer);

   // allocate the memory for the image
   if (!(bitmap->buffer = (UCHAR *)malloc(bitmap->bitmapinfoheader.biSizeImage)))
      {
      // close the file
      _lclose(file_handle);

      // return error
      return(0);
      } // end if

   // now read it in
   _lread(file_handle,bitmap->buffer,bitmap->bitmapinfoheader.biSizeImage);

   } // end if
else
if (bitmap->bitmapinfoheader.biBitCount==24)
   {
   // allocate temporary buffer to load 24 bit image
   if (!(temp_buffer = (UCHAR *)malloc(bitmap->bitmapinfoheader.biSizeImage)))
      {
      // close the file
      _lclose(file_handle);

      // return error
      return(0);
      } // end if

   // allocate final 24 bit storage buffer
   if (!(bitmap->buffer=(UCHAR *)malloc(2*bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight)))
      {
      // close the file
      _lclose(file_handle);

      // release working buffer
      free(temp_buffer);

      // return error
       return(0);
      } // end if

   // now read the file in
   _lread(file_handle,temp_buffer,bitmap->bitmapinfoheader.biSizeImage);

   // now convert each 24 bit RGB value into a 16 bit value
   for (index=0; index < bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight; index++)
       {
       // build up 16 bit color word
       USHORT color;

       // build pixel based on format of directdraw surface
       if (dd_pixel_format==DD_PIXEL_FORMAT555)
           {
           // extract RGB components (in BGR order), note the scaling
           UCHAR blue  = (temp_buffer[index*3 + 0] >> 3),
                 green = (temp_buffer[index*3 + 1] >> 3),
                 red   = (temp_buffer[index*3 + 2] >> 3);
           // use the 555 macro
           color = _RGB16BIT555(red,green,blue);
           } // end if 555
       else
       if (dd_pixel_format==DD_PIXEL_FORMAT565)
          {
          // extract RGB components (in BGR order), note the scaling
           UCHAR blue  = (temp_buffer[index*3 + 0] >> 3),
                 green = (temp_buffer[index*3 + 1] >> 2),
                 red   = (temp_buffer[index*3 + 2] >> 3);

           // use the 565 macro
           color = _RGB16BIT565(red,green,blue);

          } // end if 565

       // write color to buffer
       ((USHORT *)bitmap->buffer)[index] = color;

       } // end for index

   // finally write out the correct number of bits
   bitmap->bitmapinfoheader.biBitCount=16;

   // release working buffer
   free(temp_buffer);

   } // end if 24 bit
else
   {
   // serious problem
   return(0);

   } // end else

#if 1
// write the file info out
printf("\nfilename:%s \nsize=%d \nwidth=%d \nheight=%d \nbitsperpixel=%d \ncolors=%d \nimpcolors=%d\n",
        filename,
        bitmap->bitmapinfoheader.biSizeImage,
        bitmap->bitmapinfoheader.biWidth,
        bitmap->bitmapinfoheader.biHeight,
		bitmap->bitmapinfoheader.biBitCount,
        bitmap->bitmapinfoheader.biClrUsed,
        bitmap->bitmapinfoheader.biClrImportant);
#endif

// close the file
_lclose(file_handle);

// flip the bitmap
if (flip)
    Flip_Bitmap(bitmap->buffer,
                bitmap->bitmapinfoheader.biWidth*(bitmap->bitmapinfoheader.biBitCount/8),
                bitmap->bitmapinfoheader.biHeight);

// return success
return(1);

} // end Load_Bitmap_File

/////////////////////////////////////////////////////////////////////////

int Unload_Bitmap_File(BITMAP_FILE_PTR bitmap)
{
// this function releases all memory associated with "bitmap"
if (bitmap->buffer)
   {
   // release memory
   free(bitmap->buffer);

   // reset pointer
   bitmap->buffer = NULL;

   } // end if

// return success
return(1);

} // end Unload_Bitmap_File

/////////////////////////////////////////////////////////////////////////

int Flip_Bitmap(UCHAR *image, int bytes_per_line, int height)
{
// this function is used to flip bottom-up .BMP images

UCHAR *buffer; // used to perform the image processing
int index;     // looping index

// allocate the temporary buffer
if (!(buffer = (UCHAR *)malloc(bytes_per_line*height)))
   return(0);

// copy image to work area
memcpy(buffer,image,bytes_per_line*height);

// flip vertically
for (index=0; index < height; index++)
    memcpy(&image[((height-1) - index)*bytes_per_line],
           &buffer[index*bytes_per_line], bytes_per_line);

// release the memory
free(buffer);

// return success
return(1);

} // end Flip_Bitmap

/////////////////////////////////////////////////////////////////////////


void main(int argc, char **argv)
{
BYTE bitmap_buffer[64*64*3];       // used to buffer bitmap as its built and functions are called, can hold up to a 64x64 RGB image
BYTE output_bitmap[64*64];         // used to buffer converted bitmap, 1-byte per pixel

int x, y, px, py, tx, ty, index;   // looping vars

// intialize anything here..


// parse the command line 

// test for help request?
if (argc==2 && (!stricmp(argv[1],"-?") || !stricmp(argv[1],"-help")) || argc==1 )
    {
    // print help menu out
    printf("\nUsage BMP2DATA.exe inputfile [flags] > outputfile");
    printf("\n");
    printf("\nflags = -B -D -VGA -Axx -TW[1...255] -TH[1..255] -W[1..256] -H[1..256] -C[1..256] -STx,y -XYx,y -FX -FY -V -I -P[0|1] -?");
    printf("\n");
    printf("\n-B    = Enable 1 pixel border for templated bitmaps, default no border.");
    printf("\n-D    = Output data in hex format, default is decimal.");
    printf("\n-VGA  = Perform conversion in VGA RGB space to [xx.2.2.2] format, default is NTSC [Luma4:Chroma4] format color words.");
    printf("\n-Axx  = Alpha mapper, maps xx 8-bit color index to 0, this is useful in NTSC modes where $4F is 'black', but sprites need 0 as 'transparent'.");
    printf("\n-TWxx = Width of tile set on x-axis, default = 1.");
    printf("\n-THxx = Height of tile set on y-axis, default = 1.");
    printf("\n-Wxx  = Width of tile, default = 8 pixels.");
    printf("\n-Hxx  = Height of tile, default = 8 pixels.");
    printf("\n-Cxx  = Total count of tiles to be converted, if omitted assumed to be tw*th.");
    printf("\n-STxx,yy = Starting tile x,y indices to start scan from in scan mode, default 0,0.");
    printf("\n-XYxx,yy = Coordinate of single tile to pull from bitmap, upper left (0,0) overrides -Cxx.");
    printf("\n-FX   = Flips the output bitmap's on the X axis.");
    printf("\n-FY   = Flips the output bitmap on the Y axis.");
    printf("\n-P file = Loads in another palette bmp file, must be 192x1x256 BMP format."); 
    printf("\n-I    = Enables interactive mode.");
    printf("\n-V    = Verbose flag.");
    printf("\n-?    = Prints help.");

    exit(0);
    } // end if


// assume first parm after .exe is the inputfile, good place for more error handling
strcpy(inputfilename, argv[1]);

// now parse thru remaining command line parms and set all command line globals based on them
for (int cmd_index = 2; cmd_index < argc; cmd_index++)
    {
    // what flag is this?
    if (_strnicmp( argv[cmd_index], "-B", 2) == 0)
       {
       //printf("\nFound -B");

       // set border 
       cmd_B = 1;

       } // end if
    else if (_strnicmp( argv[cmd_index], "-D", 2) == 0)
           {
           //printf("\nFound -D");

            // set decimal output mode
            cmd_D = 1;
            } // end if
    else if (_strnicmp( argv[cmd_index], "-VGA",4) == 0)
            {
            //printf("\nFound -VGA");
            // set to vga mode
            cmd_VGA = XGS_VGA;
            } // end if        
    else if (_strnicmp( argv[cmd_index], "-A",2) == 0)
            {
            // parse out the number
            cmd_A = atoi(argv[cmd_index] + 2);
            //printf("\nFound -Axx=%d", cmd_A);
            } // end if
    else if (_strnicmp( argv[cmd_index], "-TW",3) == 0)
            {
            // parse out the number
            cmd_TW = atoi(argv[cmd_index] + 3);
            //printf("\nFound -TW=%d", cmd_TW);
            } // end if        
    else if (_strnicmp( argv[cmd_index], "-TH", 3) == 0)
            {
            // parse out the number
            cmd_TH = atoi(argv[cmd_index] + 3);
            //printf("\nFound -TH=%d", cmd_TH);

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-W",2) == 0)
            {
            // parse out the number
            cmd_W = atoi(argv[cmd_index] + 2);
            //printf("\nFound -W=%d", cmd_W);

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-H",2) == 0)
            {
            cmd_H = atoi(argv[cmd_index] + 2);
            //printf("\nFound -H=%d", cmd_H);

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-C",2) == 0)
            {
            cmd_C = atoi(argv[cmd_index] + 2);
            //printf("\nFound -C=%d", cmd_C);

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-XY",3) == 0)
            {
            // parse out the numbers in format "xx,yy"
            strcpy(cstring1, argv[cmd_index]+3); 
            strcpy(cstring2, argv[cmd_index]+3); 
        
            // find comma delimeter
            for (cindex1=0; cindex1 < strlen(cstring1); cindex1++)
                if (cstring1[cindex1] == ',')
                   break;

            // xxx,yyy
            //    |
            // cindex1

            // a little error handling
            if (cindex1 >= strlen(cstring1))
               { printf("\nParser error on -XY flag!"); exit(0); }

            // now convert both strings to numbers
            cstring1[cindex1] = 0; // holds x, null terminate
            
            cmd_X = atoi(cstring1);
            cmd_Y = atoi(cstring2+cindex1+1);

            cmd_XY = 1;
            //printf("\nFound -XY=%d, %d", cmd_X, cmd_Y);

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-ST",3) == 0)
            {
            // parse out the numbers in format "xx,yy"
            strcpy(cstring1, argv[cmd_index]+3); 
            strcpy(cstring2, argv[cmd_index]+3); 
        
            // find comma delimeter
            for (cindex1=0; cindex1 < strlen(cstring1); cindex1++)
                if (cstring1[cindex1] == ',')
                   break;

            // xxx,yyy
            //    |
            // cindex1

            // a little error handling
            if (cindex1 >= strlen(cstring1))
               { printf("\nParser error on -ST flag!"); exit(0); }

            // now convert both strings to numbers
            cstring1[cindex1] = 0; // holds x, null terminate
            
            cmd_SX = atoi(cstring1);
            cmd_SY = atoi(cstring2+cindex1+1);
            
            //printf("\nFound -ST=%d, %d", cmd_SX, cmd_SY);

            } // end if        

    else if (_strnicmp( argv[cmd_index], "-FX", 3) == 0)
            {
            cmd_FX = 1;
            //printf("\nFound -FX");

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-FY", 3) == 0)
            {
            cmd_FY = 1;
            //printf("\nFound -FY");

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-V", 2) == 0)
            {
            cmd_V = 1;
            //printf("\nEntering verbose debugging mode.");

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-I", 2) == 0)
            {
            cmd_I = 1;
            //printf("\nFound -I");

            } // end if        
    else if (_strnicmp( argv[cmd_index], "-P",2) == 0)
            {
            //cmd_P = atoi(argv[cmd_index] + 2);
            // copy filename into cmd_P
            strcpy(cmd_P, argv[cmd_index] + 2);
            //printf("\nFound -P=%d", cmd_P);

            } // end if        
    
    } // end for cmd_index

// set system palette
if (strlen(cmd_P)==0)
    {
    // load default palette for NTSC (192x1 bitmap colors left to right)
    //Load_Bitmap_File2(&palette_bitmap, "xgs_color_reference_avr.bmp"); 
    Load_Bitmap_File2(&palette_bitmap, "xgs_default_NTSC_palette_192x1.bmp");

    // scan thru RGB bytes of pixels that represent palette colors
    for (index=0; index < 192; index++)
        {
#if 0
        printf("\ncolor word %d = [%d, %d, %d]", index, palette_bitmap.buffer[index*3+0], 
                                                        palette_bitmap.buffer[index*3+1],
                                                        palette_bitmap.buffer[index*3+2] );
#endif
        // store palette information
        xgs_color_map0[ index ].peRed   = palette_bitmap.buffer[index*3+0];
        xgs_color_map0[ index ].peGreen = palette_bitmap.buffer[index*3+1];      
        xgs_color_map0[ index ].peBlue  = palette_bitmap.buffer[index*3+2];

        } // end for index

    // assign global pointer
    xgs_color_map = xgs_color_map0;
    } // end if 
 else
    {
    // load user supplied palette (working directory only)
    // todo

    xgs_color_map = xgs_color_map0;
    } // end else


// extract only root of filename
strcpy(inputfileroot, inputfilename);
for (int i=0; i < strlen(inputfileroot); i++)
    if (inputfileroot[i]=='.') { inputfileroot[i] = 0; break; }
    
// now shorten name to no longer than 12 characters since it has to be used in spin var names and there is a 30 char limit
if (strlen(inputfileroot) >= 12)
   inputfileroot[12] = 0;

//printf("\ninputfilename root/short = %s", inputfileroot);


// now load in the working bitmap we are going to extract the bitmaps from
if (!Load_Bitmap_File2(&bitmap, inputfilename))
   { printf("\nFile open error!"); exit(0); }

// print header information for XGS C/C++/ASM compliant file
printf("\n// Automated output from Bmp2Data.exe file conversion tool Version 1.0 Nurve Networks LLC\n");

// test for mode of operation

// single bitmap tile extraction?
if (cmd_XY)
   {
   // user requested a single tile
    Extract_Bitmap(cmd_X,cmd_Y,             // tile x,y
                   cmd_W,cmd_H,             // width and height of tile
                   cmd_W+cmd_B,cmd_H+cmd_B, // scan step x and y
                   bitmap_buffer, &bitmap);

    // now convert to target palette    
    Convert_Bitmap_To_Palette(bitmap_buffer,       // input bitmap in RGB format 
                              cmd_W,cmd_H,         // size of bitmap to convert
                              cmd_X,cmd_Y,         // tile indices from caller, send (0,0) if single tile
                              cmd_VGA,             // conversion mode XGS_NTSC, XGS_PAL (not supported yet), XGS_VGA
                              output_bitmap);      // output bitmap in 8-bit "data" word format ready for the XGS drivers
                                                   // each 8-bit value is in NTSC, PAL, or VGA (RGB mode)

   // return 
   exit(0);
   } // end if


// interactive mode?
if (cmd_I)
   {
   printf("\nEntering interactive mode...\n");
    while(1)
        {
        printf("\nEnter \"x y\" to scan bitmap from (enter \"-1 -1\" to exit)?");
        scanf("%d %d", &x, &y);

        if (x==-1 || y==-1) break;

        Extract_Bitmap(x,y,                     // tile x,y
                       cmd_W,cmd_H,             // width and height of tile
                       cmd_W+cmd_B,cmd_H+cmd_B, // scan step x and y
                       bitmap_buffer, &bitmap);
        

    // now convert to target palette    
    Convert_Bitmap_To_Palette(bitmap_buffer,       // input bitmap in RGB format 
                              cmd_W,cmd_H,         // size of bitmap to convert
                              x,y,               // tile indices from caller, send (0,0) if single tile
                              cmd_VGA,            // conversion mode XGS_NTSC, XGS_PAL (not supported yet), XGS_VGA
                              output_bitmap);      // output bitmap in 8-bit "data" word format ready for the XGS drivers
                                                   // each 8-bit value is in NTSC, PAL, or VGA (RGB mode)
   
        } // end while
   

 // return
 exit(0);   

 } // end if interactive mode

// if its not interactive mode and not XY single bitmap mode then user must want standard "template" scan mode, simple, just loop on tile width and height
// and end loop when we reach the "count" he requested

// initialize iterators
px = cmd_SX;
py = cmd_SY;

while(cmd_C-- > 0)
    {
    Extract_Bitmap(px,py,                // tile x,y
                cmd_W,cmd_H,             // width and height of tile
                cmd_W+cmd_B,cmd_H+cmd_B, // scan step x and y
                bitmap_buffer, &bitmap);
        

    // now convert to target palette    
    Convert_Bitmap_To_Palette(bitmap_buffer,       // input bitmap in RGB format 
                              cmd_W,cmd_H,         // size of bitmap to convert
                              px,py,               // tile indices from caller, send (0,0) if single tile
                              cmd_VGA,            // conversion mode XGS_NTSC, XGS_PAL (not supported yet), XGS_VGA
                              output_bitmap);      // output bitmap in 8-bit "data" word format ready for the XGS drivers
                                                   // each 8-bit value is in NTSC, PAL, or VGA (RGB mode)

    // move to next tile, if end of template move back left and down 1 row
    if (++px >= cmd_TW)
       { px = 0; py++; }

    } // end while

// end tile scan mode ////////////////////////////

} // end main
