///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright Nurve Networks LLC 2008
//
// Filename: CHAM_PIC_FLASH_DEMO_01.c
//
// Original Author: Andre' LaMothe
//
// Last Modified: 10.22.09
//
// Overview: This program shows the use of the Chameleon's FLASH memory. For fun, we will use "X-Modem" protocol to send
// a file from the PC using ZOC, or Absolute Telnet (something with X-Modem support). The file will be saved
// into the Chameleon's 1MB flash memory, so you can display it later, or overwrite it.
//
// You will need to hook up the USB serial port cable to the Chameleon PIC and make sure to put the serial select switch
// in the down position. Once you compile and download the program to the Chameleon PIC, you will see a simple menu display
// on both the NTSC and VGA screens.
//
// REQUIREMENTS:
//
// 1. You will need to plug in a keyboard into the Chameleon to control the program.
// 2. You will need to run an X-modem compatible uploader program on the PC with the serial USB connected to the Chameleon.
//    set the program for N81, 2400 BAUD this time.
// 3. Displays output on NTSC/VGA monitor.
//
// Some good serial programs that support a lot of features are:
//
// PuTTY           - http://www.chiark.greenend.org.uk/~sgtatham/putty/
// Zoc             - http://www.emtec.com/zoc/
// Absolute Telnet - http://www.celestialsoftware.net/telnet/
//
// There are many others, but these are pretty good. However, do not, under ANY circumstances use Windows HYPERTERMINAL -
// it is JUNK. Locks up, terribly written, single threaded, bad news.
//
// Terminal Program Setup: Once you have installed a terminal program, you will run it and set it up for communication.
// You will need to connect it to the correct serial port and then set the following settings:
//
// Data Format: "N81", no parity, 8 data bits, 1 stop bit.
// Baud Rate: 2400. Anything faster has single bit errors now and then and without any error recover, it causes the protocol to hang.
// Handshaking: None.
// X-Modem download mode with a .txt file that is smaller than 1024 bytes.
// Screen Size: 80x24 or larger, 120x50 is nice.
// Local Echo: Off.
//
// Once you have the setup, go ahead and place the program in X-Modem upload mode with a small text program ready to upload (< 1024 bytes).
// Hit RESET on the Chameleon and you should see the demo menu on the NTSC and VGA screens.
//
// Controls: PC + PC keyboard, as well as local keyboard on the Chameleon.
//
// A main menu will display, there are 2 options:
//
// 1. Xmodem receive to FLASH.
// 2. Dump FLASH to screen.
//
// Option 1 - this lets you read in a file from the PC with "XMODEM" protocol, N81, 2400. To use option 1, simple start the transfer
//            of the file on the PC, then hit 1 on the local Chameleon keyboard. The program will read and display the entire file,
//            but only save the first 1024 bytes to FLASH.
//
// Option 2 - this displays the current contents of the first 1024 bytes of FLASH memory, so you can see what you write into it.
//
// I have given you files to test with they are famous poems by various authors, the files are located in the \Source directory
// with the following names:
//
// xmodem_test_file_01.txt
// xmodem_test_file_02.txt
// xmodem_test_file_03.txt
//
// This way you don't have to hunt around for text files on your PC to test.
//
// Demo must be linked with following source files (in addition to primary file):
//
// CHAM_PIC_SYSTEM_V010.c              - System variables.
// CHAM_PIC_UART_DRV_V010.c            - UART driver.
// CHAM_PIC_I2C_SPI_DRV_V010.c         - SPI driver.
// CHAM_PIC_FLASH_DRV_V010.c           - Flash driver.
// CHAM_PIC_NTSC_DRV_V010.c            - NTSC driver.
// CHAM_PIC_VGA_DRV_V010.c             - VGA driver.
// CHAM_PIC_KEYBOARD_DRV_V010.c        - Keyboard driver.
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// include everything
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

// include local Chameleon API header files
#include "CHAM_PIC_SYSTEM_V010.h"             // system driver, you always need
#include "CHAM_PIC_UART_DRV_V010.h"           // UART driver
#include "CHAM_PIC_I2C_SPI_DRV_V010.h"        // you need this always
#include "CHAM_PIC_FLASH_DRV_V010.h"          // Flash driver
#include "CHAM_PIC_NTSC_DRV_V010.h"           // NTSC driver
#include "CHAM_PIC_KEYBOARD_DRV_V010.h"       // Keyboard driver
#include "CHAM_PIC_VGA_DRV_V010.h"            // VGA driver

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define MAX_FILE_SIZE          1024 // largest file size we can buffer

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// xmodem packet buffer for incoming files
char file_buffer[MAX_FILE_SIZE]; // buffer for incoming file

// global string building space
char sbuffer[64];

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int XModem_Receive( char *file_buffer, int buffer_size );

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int XModem_Receive( char *file_buffer, int buffer_size )
{
// receives file from an xmodem sender, standard "XMODEM" protocal as shown here:
// http://en.wikipedia.org/wiki/XMODEM
// http://www.techheap.com/communication/modems/xmodem.html
//
// function takes a pointer to the storage area of file along with the maximum size of the buffer
// if the file exceeds the buffer, the funtion WILL read the file and compute its lenght, but only the
// first buffer_size number of bytes will be written to the file_buffer.

// function returns the length of the file in bytes
//
// This is a "receiver" oriented protocal, so when you want to receive a file from the PC, start the
// transmission on the PC end FIRST, THEN run your xmodem download on this side. The PC will wait a min or 2
// before timing out, but this funtion will immediately start sending NAKs and expect the sender to respond
// I leave it to you to make this function more "robust", I wrote it in about an hour :)
//
// note: debug information only prints to NTSC currently

// xmodem protocal defines

// state machine
#define XSTATE_INIT            0
#define XSTATE_SEND_NAK        1
#define XSTATE_RECEIVE_PACKET  2
#define XSTATE_SEND_ACK        3
#define XSTATE_EOT             4

// control characters used in xmodem protocal
#define CHR_SOH                1
#define CHR_EOT                4
#define CHR_ACK                6
#define CHR_NAK               21
#define CHR_CAN               24
#define CHR_CTRLZ             26

// control debug output to NTSC screen, this is handy if you try to add more features to the function
#define DEBUG_XMODEM           1  // 0 - no output at all, 1 - display file data and file size, 2 - display everything

int xstate, xstate_running, xbytes_read = 0, index;
unsigned char ch, xchecksum, xchecksum_computed, xpacketnum, xpacketnumn;
int ret;

// start off in initialization state
xstate = XSTATE_INIT;
xstate_running = 1;

// main loop
while(xstate_running)
{
// begin x-mode state machine
switch( xstate )
    {
    case XSTATE_INIT:            // initialize all the variables and data structures for the xmodem transmission
        {

#if DEBUG_XMODEM >= 2

        NTSC_Term_Print("XSTATE_INIT");
        NTSC_Term_Char( 0x0D );
#endif

        // reset bytes read counter, etc.
        xbytes_read = 0;

        // transition to NAK state to initiate xmode transfer
        xstate = XSTATE_SEND_NAK;

        } break;

    case XSTATE_SEND_NAK:        // start the transmission, or resend current packet
        {
#if DEBUG_XMODEM >= 2
        NTSC_Term_Print("XSTATE_SEND_NAK");
        NTSC_Term_Char( 0x0D );

        NTSC_Term_Print("S:NAK");
        NTSC_Term_Char( 0x0D );
#endif
        // send NAK and wait
        UART_putchar( CHR_NAK );

        // wait for a packet..
        xstate = XSTATE_RECEIVE_PACKET;

        } break;

    case XSTATE_RECEIVE_PACKET:  // receives the packet
        {
#if DEBUG_XMODEM >= 2
        NTSC_Term_Print("XSTATE_RECEIVE_PACKET");
        NTSC_Term_Char( 0x0D );
#endif
        // wait for result
        while (1) { ret = UART_getchar(&ch); if (ch == CHR_SOH || ch == CHR_EOT) break; } // end while

#if DEBUG_XMODEM >= 2
        sprintf(sbuffer,"Looking for SOT/EOT: %d %c",ch, 0x0D );
        NTSC_Term_Print( sbuffer );
#endif

        // we have the first byte, test for SOH or EOT
        if (ch == CHR_SOH)
            {
#if DEBUG_XMODEM >= 2
            NTSC_Term_Print("R:SOH");
            NTSC_Term_Char( 0x0D );
#endif
            } // end if SOH
        else if (ch == CHR_EOT)
            {
#if DEBUG_XMODEM >= 2
            NTSC_Term_Print("R:EOT");
            NTSC_Term_Char( 0x0D );
#endif
            xstate = XSTATE_EOT;
            break;

            } // end if

        // get packet number
        while (1) { ret = UART_getchar(&ch); if (ret != 0 || ch == CHR_CTRLZ) break; } // end while

        xpacketnum = ch;

#if DEBUG_XMODEM >= 2
        sprintf(sbuffer,"Packet # %d\r",xpacketnum );
        NTSC_Term_Print( sbuffer );
#endif

        // get packet number complement
        while (1) { ret = UART_getchar(&ch); if (ret != 0 || ch == CHR_CTRLZ) break; } // end while
        xpacketnumn = ch;

#if DEBUG_XMODEM >= 2
        sprintf(sbuffer,"Packetn (255-packet)# %d\r",xpacketnumn );
        NTSC_Term_Print( sbuffer );
#endif
        // now get the packet
        xchecksum_computed = 0;

#if DEBUG_XMODEM >= 2
        NTSC_Term_Print("DATA:");
        NTSC_Term_Char( 0x0D );
#endif
        for (index = 0; index < 128; index++)
            {
            // wait for next byte
            while (1) { ret = UART_getchar(&ch); if (ret != 0 || ch == CHR_CTRLZ) break; } // end while

            // insert into buffer (if still room?)
            if ( (index + (xpacketnum-1)*128) < buffer_size )
              file_buffer[ index + (xpacketnum-1)*128 ] = ch;

            // update checksum
            xchecksum_computed += ch;


#if DEBUG_XMODEM >= 1
            // test for printable
            if ( (ch >= 32 && ch <= 127) || ch==0x0D)
              {
              NTSC_Term_Char( ch );
              VGA_Term_Char( ch );
              } // end if printable
            else // non-printable character
              {
              // print something for place holder?
              //NTSC_Term_Char( '.' );
              //VGA_Term_Char( '.' );
              } // end else non-printable
#endif

            } // end for index

        // add bytes read
        xbytes_read += 128;

        // read the checksum
        while (1) { ret = UART_getchar(&ch); if (ret != 0 || ch == CHR_CTRLZ) break; } // end while
        xchecksum = ch;

#if DEBUG_XMODEM >= 2
        NTSC_Term_Char( 0x0D );
        sprintf(sbuffer,"Checksum: Sent=%d, Calc=%d\r",xchecksum, xchecksum_computed );
        NTSC_Term_Print( sbuffer );
        NTSC_Term_Char( 0x0D );
#endif

        // send the ACK (test if packet was good later)
        xstate = XSTATE_SEND_ACK;

        } break;

    case XSTATE_SEND_ACK:        // packet was good, please send another?
        {
#if DEBUG_XMODEM >= 2
        NTSC_Term_Print("XSTATE_SEND_ACK");
        NTSC_Term_Char( 0x0D );

        NTSC_Term_Print("S:ACK");
        NTSC_Term_Char( 0x0D );
#endif
        DELAY_MS ( 100 );

        // send ACK and wait
        UART_putchar( CHR_ACK );

        DELAY_MS ( 100 );

        // wait for a packet..
        xstate = XSTATE_RECEIVE_PACKET;

        } break;

    case XSTATE_EOT:            // possible end of transmission
        {
#if DEBUG_XMODEM >= 2
        NTSC_Term_Print("XSTATE_EOT");
        NTSC_Term_Char( 0x0D );

        NTSC_Term_Print("S:NAK");
        NTSC_Term_Char( 0x0D );
#endif
        // send NAK, wait for another EOT
        UART_putchar( CHR_NAK );

        // wait for result
        while (1) { ret = UART_getchar(&ch); if (ret != 0 || ch == CHR_CTRLZ) break; } // end while

        // test for final EOT
        if (ch == CHR_EOT )
            {
            // final EOT, send ACK, finish
            UART_putchar ( CHR_ACK );

#if DEBUG_XMODEM >= 2
            NTSC_Term_Print("S:ACK");
            NTSC_Term_Char( 0x0D );
#endif
            // exit state machine
            xstate_running = 0;

            } // end if

        } break;

    default: break;

    } // end switch

} // end while

#if DEBUG_XMODEM >= 1
sprintf(sbuffer,"\rDownload Complete.\r%d Blocks.\rFile Size = %d.\r", xbytes_read/128, xbytes_read );
NTSC_Term_Print( sbuffer );
VGA_Term_Print( sbuffer );
#endif

// return bytes read
return ( xbytes_read );

} // end XModem_Receive

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
int index, ch, key;   // looping, input chars
int bytes_received;   // bytes read by xmodem receive

// Always call ConfigureClock first
SYS_ConfigureClock(MAX_FCY_RATE);

// initialize SPI
SPI_Init();

// initialize UART to slow speed, so no xmodem errors, my function has no error correction yet!
UART_Init( 2400 );

// give Propeller driver time to boot
DELAY_MS( 2500 );

// load keyboard driver (this driver is already loaded by default, but this call can't hurt)
Keyboard_Load();

// clear screens
NTSC_ClearScreen();
NTSC_Color(0);

VGA_ClearScreen();
VGA_Color(0);


// pop up header
NTSC_Term_Print(" Chameleon FLASH Memory Demo");
NTSC_Term_Char(0x0D);
VGA_Term_Print(" Chameleon FLASH Memory Demo");
VGA_Term_Char(0x0D);



// main event loop
while(1)
    {
    NTSC_Term_Char(0x0D);
    NTSC_Term_Print(" Main Menu");
    NTSC_Term_Char(0x0D);
    NTSC_Term_Char(0x0D);
    NTSC_Term_Print(" 1. XModem receive to FLASH.");
    NTSC_Term_Char(0x0D);
    NTSC_Term_Print(" 2. Dump FLASH to screen.");
    NTSC_Term_Char(0x0D);
    NTSC_Term_Char(0x0D);
    NTSC_Term_Print(" Selection (1,2)?");

    VGA_Term_Char(0x0D);
    VGA_Term_Print(" Main Menu");
    VGA_Term_Char(0x0D);
    VGA_Term_Char(0x0D);
    VGA_Term_Print(" 1. XModem receive to FLASH.");
    VGA_Term_Char(0x0D);
    VGA_Term_Print(" 2. Dump FLASH to screen.");
    VGA_Term_Char(0x0D);
    VGA_Term_Char(0x0D);
    VGA_Term_Print(" Selection (1,2)?");

     // read next key, return 0 (null if no key), wait for key press
     while ( (key = Keyboard_Key()) == 0);

     // print to screen (better be printable!)
     if ( key >= 32 && key <= 127)
       NTSC_Term_Char( key );

     // which selection?
     if (key == '1')
        {
        // give user a moment to prepare
        NTSC_Term_Char(0x0D);
        NTSC_Term_Print(" X-Modem Starting in 3 secs...");
        NTSC_Term_Char( 0x0D );
        VGA_Term_Char(0x0D);
        VGA_Term_Print(" X-Modem Starting in 3 secs...");
        VGA_Term_Char( 0x0D );
        DELAY_MS( 1000 );

        NTSC_Term_Print(" X-Modem Starting in 2 secs...");
        NTSC_Term_Char( 0x0D );
        VGA_Term_Print(" X-Modem Starting in 2 secs...");
        VGA_Term_Char( 0x0D );

        DELAY_MS( 1000 );

        NTSC_Term_Print(" X-Modem Starting in 1 secs...");
        NTSC_Term_Char( 0x0D );
        VGA_Term_Print(" X-Modem Starting in 1 secs...");
        VGA_Term_Char( 0x0D );

        DELAY_MS( 1000 );

        // receive file into buffer
        bytes_received = XModem_Receive( file_buffer, MAX_FILE_SIZE );

        // open the 1MB FLASH memory up and unprotect the sectors
        NTSC_Term_Char(0x0D);
        NTSC_Term_Print(" Opening FLASH memory...");
        NTSC_Term_Char( 0x0D );

        VGA_Term_Char(0x0D);
        VGA_Term_Print(" Opening FLASH memory...");
        VGA_Term_Char( 0x0D );

        Flash_Open();

        // write file to FLASH...

        // step 1: erase the first 4K block (the function takes the 4K block number)
        NTSC_Term_Print(" Erasing first 4K of FLASH...");
        NTSC_Term_Char( 0x0D );

        VGA_Term_Print(" Erasing first 4K of FLASH...");
        VGA_Term_Char( 0x0D );

        Flash_Erase( FLASH_ADDR_TO_BLOCK( 0 ) );

        // now write the bytes to the FLASH starting at address 0...
        sprintf(sbuffer," Writing %d bytes to FLASH...",bytes_received );
        NTSC_Term_Print( sbuffer );
        VGA_Term_Print( sbuffer );

        Flash_Write(0, (unsigned char *)file_buffer, bytes_received);

        // close the 1MB FLASH memory and protect the sectors
        NTSC_Term_Char( 0x0D );
        NTSC_Term_Print(" Closing FLASH memory...");
        NTSC_Term_Char( 0x0D );

        VGA_Term_Char( 0x0D );
        VGA_Term_Print(" Closing FLASH memory...");
        VGA_Term_Char( 0x0D );

        Flash_Close();

        } // end if
    else
    if (key == '2')
        {
        // open the 1MB FLASH memory up and unprotect the sectors
        NTSC_Term_Char(0x0D);
        NTSC_Term_Print(" Opening FLASH memory...");
        NTSC_Term_Char( 0x0D );

        VGA_Term_Char(0x0D);
        VGA_Term_Print(" Opening FLASH memory...");
        VGA_Term_Char( 0x0D );

        Flash_Open();

        // read and print the first MAX_FILE_SIZE bytes of the FLASH memory, non-printable character will print at dots '.'
        NTSC_Term_Print(" Reading data from FLASH...");
        NTSC_Term_Char( 0x0D );

        VGA_Term_Print(" Reading data from FLASH...");
        VGA_Term_Char( 0x0D );

        Flash_Read(0, (unsigned char *)file_buffer, MAX_FILE_SIZE );

        // print the data to the screen
        for (index = 0; index < MAX_FILE_SIZE; index++)
            {
            // get next character from buffer
            ch = file_buffer[ index ];

            // test for printable / non-printable
            if ( (ch >= 32 && ch <= 127) || ch == 13)
              {
              NTSC_Term_Char( ch );
              VGA_Term_Char( ch );
              }
            else
              {
              NTSC_Term_Char( '.' );
              VGA_Term_Char( '.' );
              }

            // pause every 128 characters
            if ( (index % 128) == 127)
              while ( (key = Keyboard_Key()) == 0);

            } // end for index

        // close the 1MB FLASH memory and protect the sectors
        NTSC_Term_Char( 0x0D );
        NTSC_Term_Print(" Closing FLASH memory...");
        NTSC_Term_Char( 0x0D );

        VGA_Term_Char( 0x0D );
        VGA_Term_Print(" Closing FLASH memory...");
        VGA_Term_Char( 0x0D );

        Flash_Close();

        } // end if
    else
        {
        NTSC_Term_Char( 0x0D );
        NTSC_Term_Print(" Invalid Selection!");
        NTSC_Term_Char( 0x0D );

        VGA_Term_Char( 0x0D );
        VGA_Term_Print(" Invalid Selection!");
        VGA_Term_Char( 0x0D );

        } // end else

    } // end while

// return to caller (like there is a caller, no operating system!)
return(0);

} // end main





