///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright Nurve Networks LLC 2008
//
// Filename: CHAM_PIC_GUESS_NUMBER_01.c
//
// Original Author: Andre' LaMothe
//
// Last Modified: 10.20.09
//
// Overview: This program shows the use of the Chameleon PIC UART and serial communication to the PC. You will need to hook
// up the USB serial port cable to the Chameleon PIC and make sure to put the serial select switch in the down position.
// Once you compile and download the program to the Chameleon PIC, then you will need a terminal program
// running on the PC in VT100 mode. Some good serial programs that support a lot of features are:
//
// PuTTY           - http://www.chiark.greenend.org.uk/~sgtatham/putty/
// Zoc             - http://www.emtec.com/zoc/
// Absolute Telnet - http://www.celestialsoftware.net/telnet/
//
// There are many others, but these are pretty good. However, do not, under ANY circumstances use Windows HYPERTERMINAL -
// it is JUNK. Locks up, terriably written, single threaded, bad news.
//
// Terminal Program Setup: Once you have installed a terminal program, you will run it and set it up for communication.
// You will need to connect it to the correct serial port and then set the following settings:
//
// Data Format: "N81", no parity, 8 data bits, 1 stop bit.
// Baud Rate: 115200.
// Handshaking: None.
// Terminal Mode: VT100.
// Screen Size: 80x24 or larger, 120x50 is nice.
// Local Echo: Off.
//
// Once you have the setup, go ahead and place the program in terminal mode, hit RESET on the Chameleon and you should see
// the demo start.
//
// Controls: PC + PC keyboard
//
//
// Demo must be linked with following source files (in addition to primary file):
//
// CHAM_PIC_SYSTEM_V010.c              - System variables.
// CHAM_PIC_UART_DRV_V010.c            - UART driver.
// CHAM_PIC_I2C_SPI_DRV_V010.c         - SPI driver.
// CHAM_PIC_SOUND_DRV_V010.c           - Sound driver.
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Include C standard headers
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

// include local Chameleon API header files
#include "CHAM_PIC_SYSTEM_V010.h"
#include "CHAM_PIC_UART_DRV_V010.h"
#include "CHAM_PIC_SOUND_DRV_V010.h"
#include "CHAM_PIC_I2C_SPI_DRV_V010.h"        // you need this always

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define IsNum(ch) 		(ch >= 30 && ch <= 39)
#define IsAlpha(ch) 	((ch >= 65 && ch <= 90) || (ch >= 97 && ch <= 122))
#define IsAlphaUpr(ch)	(ch >= 65 && ch <= 90)
#define IsAlphaLwr(ch)	(ch >= 97 && ch <= 122)
#define IsAlphaNum(ch) 	{IsNum(ch) || IsAlpha(ch))

// Generates a random number in range x->y inclusive
#define RAND_RANGE(x,y) ( ( x ) + (rand()%( ( y ) - ( x )+1)))


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// .X...X.XXXXX.XXXXX...XXXX..XXXXX.XXX.XXX.XXXX.XXXXX.
// .X...X.X...X.X..XX...X.....X...X.X..X..X.X....X.....
// .X...X.XXXXX.XXXX....X..XX.XXXXX.X.....X.XXX..XXXXX.
// .X.X.X.X...X.X..X....X...X.X...X.X.....X.X........X.
// .XXXXX.X...X.X...X...XXXXX.X...X.X.....X.XXXX.XXXXX.

// the title screen bitmap as a matrix of byte, stored in FLASH for fun, then to render we will convert to
// characters: BLOCK and SPACE
// notice the use of "const" to store into flash memory
// size of banner 52x5
const unsigned char title_string_flash[] = {

0,1,0,0,0,1,0,1,1,1,1,1,0,1,1,1,1,1,0,0,0,1,1,1,1,0,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,1,0,1,1,1,1,1,0,
0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,1,1,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,1,0,0,1,0,1,0,0,0,0,1,0,0,0,0,0,
0,1,0,0,0,1,0,1,1,1,1,1,0,1,1,1,1,0,0,0,0,1,0,0,1,1,0,1,1,1,1,1,0,1,0,0,0,0,0,1,0,1,1,1,0,0,1,1,1,1,1,0,
0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,0,1,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,1,0,
0,1,1,1,1,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,1,1,1,1,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,1,1,1,0,1,1,1,1,1,0,
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void _strupr(char *CmdStr)
{
	if(CmdStr == NULL) return;
	while(*CmdStr)
	{
		if(IsAlphaLwr(*CmdStr))
			*CmdStr -= 32;
		CmdStr++;
	}
}



int Get_String( char *cmd_buff )
{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// this function waits for the user to input a string on from the UART, it supports some
// editing such as backspace. When the user hits "return" the function returns with the string
//
// INPUTS:
//          *buffer  = pointer to string storage where the raw ASCII string from the UART is placed
//
// OUTPUTS:
//          returns the length of the string
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned char c;
int cmd_buff_index = -1;


// enter infinite while until valid string is entered
while(1)
{
// process next character
if ( (UART_getchar(&c)) != 0 )
    {
    // special characters
    switch (c)
    {
    case 0x0D: // carriage return
    case 0x0A: // line feed
        {
        cmd_buff[++cmd_buff_index] = 0;

        // any command to process?
        if (cmd_buff_index > 0)
           {
            // process command...

            // upcase the string to normalize tokens
            _strupr( cmd_buff );

            // string has been entered and tokenized, now return number of tokens
            return( cmd_buff_index );
            } // end if

        // empty string
        return (0);

        } break;

    case 0x08: // back space
         {
         if (cmd_buff_index >=0 )
            {
            UART_putchar( c );
            cmd_buff_index--;
            } // end if
         } break;

    default: // otherwise normal character, insert into command string and echo
        {
        UART_putchar( c );
        cmd_buff[++cmd_buff_index] = c;
        } break;

    } // end switch

  } // end if

} // end while

} // end Get_String


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
int c;
long index;

long number, guess;

int num_tokens = 0;
char *token_ptr = NULL;
int str_len;

char cmd_buff[64];
char *tokens[8];


// Always call ConfigureClock first
SYS_ConfigureClock(MAX_FCY_RATE);

// initialize SPI
SPI_Init();

// initialize UART
UART_Init( 115200 );

// give Propeller driver time to boot
DELAY_MS( 2500 );

// clear the screen and home cursor
UART_vt100ClearScreen();

// print header
UART_printf("\r\nChameleon WarGames (C) 2019\r\n");
DELAY_MS( 2500 );

UART_printf("\r\nentering loop..\r\n");

// draw "WAR GAMES" with characters from FLASH
for (index = 0; index < 52*5; index++)
    {
    // read the image from FLASH memory with special function call that uses "LPM" ASM instruction to read FLASH
    c = title_string_flash[ index ] ;

    // based on data convert to character and send to UART
    if (c == 0)
       {
       Sound_Play(0, 1000, 1);
       UART_printf(" ");
       }
    else
    if (c == 1)
       {
       Sound_Play(0, 2000, 1);
       UART_printf("#");
       } // end if

    if ( (index % 52) == 0)
       UART_printf("\r\n");

    DELAY_MS(10);

    } // end for

Sound_Stop(0);

// print some fun text to make it seem like a smart computer
UART_printf("\r\n\r\nHello Dr. Falkner...\r\n");
DELAY_MS(2000);

UART_printf("\r\nJust a moment while I initialize my systems:");
DELAY_MS(1000);

UART_printf("\r\nHolographic Memory:");
Sound_Play(0, 500, 1);
DELAY_MS(1000);
UART_printf("OK");

UART_printf("\r\nMultiprocessor Sub-Systems:");
Sound_Play(0, 500, 1);
DELAY_MS(1000);
UART_printf("OK");

UART_printf("\r\nQuantum Telecom Links:");
Sound_Play(0, 500, 1);
DELAY_MS(1000);
UART_printf("OK");

UART_printf("\r\nLoading Autonomous Systems");
Sound_Play(0, 500, 1);

for (index=0; index < 10; index++)
    {
    UART_printf(".");
    DELAY_MS(250);
    } // end for

Sound_Play(0, 1000, 3);

UART_printf("\r\n\r\nThat's much better.\r\n");

// enter main game loop
while(1)
    {
    UART_printf("\r\nWould you like to play a game (yes or no)?");

    // get the string from the UART
    str_len = Get_String( cmd_buff );

    if ( str_len > 0 && (!strcmp( cmd_buff, "YES") || !strcmp( cmd_buff, "Y")) )
       {
       UART_printf("\r\n\r\nExcellent, let's play 'Guess My Number'...\r\n");
       DELAY_MS(1000);
       UART_printf("\r\nI am thinking of a number from 1 to 100, try and guess.\r\n");
       DELAY_MS(1000);
       UART_printf("\r\nOk, ready...\r\n");

       // get random number
       number = RAND_RANGE(1, 100);

       // enter guessing loop
       while(1)
            {
            UART_printf("\r\nYour Guess?");

            // get the string from the UART
            Get_String( cmd_buff );


            // initialize command processor
            num_tokens = 0;

            // initialize token processor
            token_ptr = strtok( cmd_buff, "=, ");

            while( token_ptr )
                 {
                 // insert token into token array and inc number of tokens found
                 tokens[ num_tokens++ ] = token_ptr;
                 token_ptr = strtok( NULL, "=, ");
                 } // end while

            // convert string to number
            guess = atoi( tokens[0] );

            // conditional logic
            if (guess > number)
               UART_printf("\r\n\r\nToo high, your going to have to do better...\r\n");
            else
            if (guess < number)
               UART_printf("\r\n\r\nToo low, this is getting embarasing for you...\r\n");
            else
               {
               UART_printf("\r\n\r\nThat's correct! Maybe there is hope for carbon based lifeforms!\r\n");
               break;
               } // end else

            } // end while

       } // end if
    else
    if (  str_len > 0 && (!strcmp( cmd_buff, "NO") || !strcmp( cmd_buff, "N")) )
       {
       UART_printf("\r\n\r\nWell fine, I am now running a holographic worm on your computer's memory...");

       // and lock up the Chameleon for fun
       while(1);

       } // end if
    else
       {
       // user is typing bad commands, let him know!
       switch( RAND_RANGE(0,3) )
            {
            case 0: UART_printf("\r\nI don't understand..please try again.\r\n"); break;
            case 1: UART_printf("\r\nCan you please answer the question again?\r\n"); break;
            case 2: UART_printf("\r\nI am starting to loose my patience Dr. Falkner...\r\n"); break;
            case 3: UART_printf("\r\nI can see why SkyNet wants to exterminate your kind.\r\n"); break;
            } // end switch

       } // end else

    } // end while

// return to caller (like there is a caller, no operating system!)
return(0);

} // end main





