///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright Nurve Networks LLC 2009
//
// Filename: CHAM_PIC_PROPPORT_DEMO_01.c
//
// Original Author: Andre' LaMothe
//
// Last Modified: 10.20.09
//
// Description: Demo of the Propeller Local Port
//
// Overview: Simply blinks an LED plugged into the Prop Port, the trick is to set one pin high (3.3V) and the other
// low (ground) which will turn the LED on, or with a bi-color LED, it will turn the LED on with one of the colors.
//
// Normally, with a standard LED, you would have to insert the LED into the Prop Port so that the
// LED is in a standard orientation such as the +ANODE (short pin inside LED) to the left of the port at
// P7 and the -CATHODE (long/large pin inside LED) to the right into Pin 6 for example. Then we could set pin 7 HIGH
// and pin 6 LOW and the LED would turn on as shown below:
//    ____
//   /    \
//  |      |
//  |   __ |
//  | | --||
//  --|---|-
//    |   |
//    |   |
//    |    |     Propeller Port
//  --|----|---------------------------------
//  | |  | |  |    |    |    |    |    |    |
//  | P7 | P6 | P5 | P4 | P3 | P2 | P1 | P0 |
//  -----------------------------------------
//    ^    ^
//    |    |
//  Anode Cathode
//   (+)  (-)
//
//
// However, if you have a bi-color LED, you can insert it either way and it will turn, thus, the program works with normal
// LEDs or bi-color, and in fact, it will toggle the color between RED/GREEN (or whatever the bi-color LED is) or ON/OFF
// for a standard LED. And the code toggles all the port pins, so you can actually put the LED into any two port pins and it
// will work!
//
// Note: normally you want to put a current limiting resistor in series with LEDs to make sure they don't consume too much
// current and or burn out, but since the propeller port is 3.3V, it runs the LED at a nominal current that is safe.
//
// Chameleon AVR version of program nearly identical
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// include headers required for demo
#include "CHAM_PIC_SYSTEM_V010.h"             // you need this one always
#include "CHAM_PIC_I2C_SPI_DRV_V010.h"        // you need this always
#include "CHAM_PIC_NTSC_DRV_V010.h"           // needed for NTSC driver
#include "CHAM_PIC_PROP_PORT_DRV_V010.h"      // needed for prop port

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
// Always call ConfigureClock first
SYS_ConfigureClock(MAX_FCY_RATE);

// initialize SPI
SPI_Init();


// give Prop a moment before sending it commands, the boot process is about a second
// if you like you can speed it up, by removing the Prop driver's "LED blink" sequence in the driver
// or speed it up, but commands that are sent before the Prop driver is done booting will be ignored.
// therefore, if you have a lot of set up work to do, then you don't require a delay, but if you jump right
// into commands, then you need a good 1.5 - 2 second delay
DELAY_MS( 2500 );

// clear screens
NTSC_ClearScreen();
NTSC_Color(0);

// set direction of prop port so we have all bits output
PropPort_SetDir( 0b11111111 ); // 1 is output, 0 input

// enter infinite loop...
while(1)
     {
     // print on NTSC terminal screen
     NTSC_Term_Print("Port Data ON [+ - + - + - + -]");
     NTSC_Term_Char( 0x0D );

     // turn on the LED pin 7 = HIGH, pin 6 = LOW, repeat pattern down port.
     PropPort_Write(0b10101010);

     // slow things down a bit, so we can read the text!
     DELAY_MS(250);

     // print on NTSC terminal screen
     NTSC_Term_Print("Port Data OFF[- + - + - + - +]");
     NTSC_Term_Char( 0x0D );

     // turn off the LED pin 7 = LOW, pin 6 = HIGH, repeat pattern down port.
     PropPort_Write(0b01010101);

     // slow things down a bit, so we can read the text!
     DELAY_MS(250);

     } // end while

} // end main
