///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright Nurve Networks LLC 2008
//
// Filename: CHAM_PIC_SYSTEM_V010.C
//
// Original Author: Joshua Hintze
//
// Last Modified: 8.31.08
//
// Description: System library file, it includes all our common types, macros, functions that are used
//				throughout the Chameleon library.
//
//
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// First and foremost, include the chip definition header
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


// include local Chameleon API header files
#include "CHAM_PIC_SYSTEM_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CHIP CONFIG SETTINGS - (FUSE BITS)  - MUST BE DEFINED ONLY ONCE IN CODE
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

_FICD(ICS_PGD3 & JTAGEN_OFF);		// Jtag off so we can use those pins, ISP on port 3
_FWDT(FWDTEN_OFF & WINDIS_OFF); 	// Disable Watchdog timer and windowed watchdog timer
_FOSC(FCKSM_CSECMD & OSCIOFNC_OFF & POSCMD_XT & IOL1WAY_OFF) // Clock monitor: disabled switch: enabled, Using XT oscillator (3-10 MHz), Allow multiple remaps
_FOSCSEL(FNOSC_FRC & IESO_OFF)		// Start with the fast rc until we switch to PLL, two speed oscillator startup disabled


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MACROS
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Global functional instruction rate. This is the instruction rate, not the clock rate. Fcy = 1/2 * Fclock
unsigned long g_FCY = MAX_FCY_RATE;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Configures the clock PLL multiplier, this function will use the #define FCY above for the clock rate
void SYS_ConfigureClock(unsigned long FCY)
{
	// Save off the clcok rate
	g_FCY = FCY;

	// To properly configure the PLL we need to switch between two clocks. First
	// start off with a Fast RC clock rate. Then modify the PLLPre,PLLPost,PLLDiv registers,
	// to set up the appropriate new clock rate. Then issue a clock switch.

	// Check for NTSC
	if(g_FCY == MAX_FCY_RATE)
	{
		// Configures the clock to 16 * Input (5 MHz) = 80 MHz or Fcy = 40 MHz
		// General Equation for PLL setup = Fosc = Fin ( M / (N1 * N2) )
		// M = PLLDIV = 64
		// N1 = PLLPRE = 2
		// N2 = PLLPOST = 2

		CLKDIVbits.PLLPRE = 0; 			// 0 = 2, 1 = 3, 2 = 4, etc
		CLKDIVbits.PLLPOST = 0; 		// 0 = 2
		PLLFBD = 62;					// PLL feedback divisor where 0 = 2, 1 = 3, etc
	}
	else // Must be VGA (In the future the user could add more speeds in here)
	{
		// Return since we didn't find a matching frequency
		return;
	}


	// We need to initiate a clock switch. This is time critical code so we will use some special built-in
	// functions to help us out and unlocks/relocks these registers.
	__builtin_write_OSCCONH(0x03);	// Select the External Clock with PLL
	__builtin_write_OSCCONL(0x01);	// Initiate the switch by setting the switch enable bit

	// Wait for Clock Switch to occur
	while (OSCCONbits.COSC != 0b011);

	// Wait for PLL to lock
	while (OSCCONbits.LOCK != 1);

}




